//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"context"
	"errors"
	"net/http"
	"sync"
	"time"
)

const (
	defaultAge        = 24 * time.Hour
	SessionCookieName = "aura-tank-session"
)

type Session struct {
	Expires  time.Time `json:"expires"`
	Username string    `json:"username"`
	IsAdmin  bool      `json:"is-admin"`
	Shows    []string  `json:"shows"`
	oidc     *OIDCSession
}

var (
	anonAllowNone = &Session{Username: "anonymous", IsAdmin: false, Shows: []string{}}
	anonAllowAll  = &Session{Username: "anonymous", IsAdmin: true, Shows: []string{}}
)

func (s *Session) Expired() bool {
	return s.Expires.Before(time.Now())
}

func getSessionFromCookie(r *http.Request) (string, *Session) {
	sc, err := r.Cookie(SessionCookieName)
	if sc == nil && err != nil {
		return "", nil
	}

	sid := sc.Value
	s := auth.sessions.get(sid)
	if s == nil {
		return "", nil
	}
	if s.Expired() {
		auth.sessions.remove(sid)
		return "", nil
	}

	return sid, s
}

func setSessionCookie(w http.ResponseWriter, sid string, expires time.Time) {
	// TODO: make cookie settings more secure!
	sc := &http.Cookie{Name: SessionCookieName}
	sc.Value = sid
	sc.Expires = expires
	sc.Path = "/"
	http.SetCookie(w, sc)
}

func invalidateSessionCookie(w http.ResponseWriter) {
	// TODO: this needs to improve
	sc := &http.Cookie{Name: SessionCookieName, Value: "invalid", MaxAge: -1}
	http.SetCookie(w, sc)
}

func attachSessionToRequest(r *http.Request, s *Session) *http.Request {
	ctx := context.WithValue(r.Context(), sessionContextKey, s)
	return r.WithContext(ctx)
}

func SessionFromRequest(r *http.Request) (*Session, bool) {
	s, ok := r.Context().Value(sessionContextKey).(*Session)
	return s, ok
}

type SessionManager struct {
	mutex    sync.RWMutex
	maxAge   time.Duration
	sessions map[string]*Session
}

func NewSessionManager(cfg SessionsConfig) (sm *SessionManager, err error) {
	sm = &SessionManager{maxAge: defaultAge}
	if cfg.MaxAge > 0 {
		sm.maxAge = cfg.MaxAge
	}
	sm.sessions = make(map[string]*Session)
	go sm.runMaintenance()
	return
}

func (sm *SessionManager) runMaintenance() {
	t := time.NewTicker(time.Minute * 10)
	for {
		<-t.C
		sm.cleanupExpired()
	}
}

func (sm *SessionManager) insert(s *Session) (id string, err error) {
	if id, err = generateRandomString(32); err != nil {
		return
	}
	s.Expires = time.Now().Add(sm.maxAge)

	sm.mutex.Lock()
	defer sm.mutex.Unlock()
	sm.sessions[id] = s
	return
}

func (sm *SessionManager) get(id string) *Session {
	sm.mutex.RLock()
	defer sm.mutex.RUnlock()

	s, ok := sm.sessions[id]
	if !ok {
		return nil
	}
	return s
}

func (sm *SessionManager) update(id string, s *Session) error {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	if _, ok := sm.sessions[id]; !ok {
		return errors.New("session not found.")
	}
	sm.sessions[id] = s
	return nil
}

func (sm *SessionManager) remove(id string) {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	delete(sm.sessions, id)
}

func (sm *SessionManager) cleanupExpired() {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	for id, s := range sm.sessions {
		if s.Expired() {
			delete(sm.sessions, id)
		}
	}
}
